;UT4 as typed from COSMAC Design Book, just code not most comment
;modified slightly for modern 1802 assemblers but code the same
;HRJ Mar 19 2014
;corrections thanks to Robert Coward July 2018 marked 7/18
;
;UI4, at initiation, stores all registers
;between 8C00 and 8C1F if it finds RAM tehre
;but R0, R1, and R4.1 are clobbered)

PTER	equ	0	;auxiliary for main routine
CL	equ	1	;clobbered
ST	equ	2	;stack pointer only 
			;ref to RAM
SUB	equ	3	;subroutine PC
PC	equ	5	;main program counter
SWITCH equ	CL	;distinguishes between ?M and !M
DELAY	equ	0CH	;delay routine program counter
ASL	equ	0DH	;hex assy reg on input,
			;aux for hex output
CENTER equ	ASL	;used to count output bytes
AUX	equ	0EH	;.1 holds bit-time constant
CHAR	equ	0FH	;.1 holds I/O byte
;
;
;calls into ROM monitor
;813E	READ
;813B	READAH
;8140 TTYRED
;819C	TYPE5D
;81A0	TYPE5
;81A2	TYPE6
;81A4	TYPE
;81AE	TYPE2
;80FE	TIMALC
;80EF	DELAY
;
;
; register equivalents for A18
R0		EQU	0
R1		EQU	1
R2		EQU	2
R3		EQU	3
R4		EQU	4
R5		EQU	5
R6		EQU	6
R7		EQU	7
R8		EQU	8
R9		EQU	9
RA		EQU	10
RB		EQU	11
RC		EQU	12
RD		EQU	13
RE		EQU	14
RF		EQU	15
;
; enter in R0
	org 8000H
	NOP
	LDI HIGH UT4
	PHI	R0	;set PC while finger is on
;the followin writes reg contents into 8C00-1F if it exists
;8BFE is assumed not to be RAM else routine overruns
	LDI 	8CH
	PHI	CL	;CL is clobbered by this routine
	LDI	1EH
	PLO	CL	;set up where RF.0 is to go minus 1
	LDI	0A0H
	PHI	R4	;R4.1 stores a modified instruction
	SEX	CL
LOOP2: LDI	0D0H
	STR	CL	;set up SEP instruction for return
	XOR		;check that it wrote
	BNZ	UT4
	DEC 	CL	;prepare for modified instruction
	GHI	R4
	ADI	70H	;see if it is in the 90's
	BDF	$+04
	ADI	21H	;if no, 8N becomes 9N
	ADI	7FH	;if yes, 9N becomes 8(N-1)
	PHI	R4
	STR	CL	;set modified instr into RAM
	XOR		;check that it wrote
	BNZ	UT4
	SEP	CL	;go to execute instru (80-9F)
	STR	CL	;store result in RAM
	DEC	CL
	DEC	CL	;back up for next byte
	BR	LOOP2
;
UT4:	GHI	R0
	PHI	PC
	PHI	SUB	;80-PC.1 and SUB.1
	LDI	LOW UT4A
	PLO	PC
	SEP	PC
UT4A:	SEX	PC
	DIS	
	DB	55H	;note PC=5 assumed here!
	OUT	1
	DB	01H	;select RCA group
	LDI	low TIMALC
	PLO	SUB	;read one to set timer
	SEP	SUB
;initiation now done
START: LDI	LOW TYPE5D
	PLO	SUB
	SEP	SUB
	DB	0DH	;CR = carriage return
ST2:	SEP	SUB
	DB	0AH	;LF = line feed
	SEP	SUB
	DB	2AH	;* as prompt char
IGNORE:
	LDI	00H
	PLO	ASL
	PHI	ASL	;prepare to input hex digits
			;clear ASL
	LDI	low READAH
	PLO	SUB
	SEP	SUB	;input command
	XRI	24H	;is it $ ?
	BZ	DOLLAR
	XRI	05H	;is it $ ?
	PLO	SWITCH ; and save result
	LSZ		;equiv to BR RDARGS
	XRI	1EH	;is it ? ?
			;test with $ xor ! xor ?
	BNZ	IGNORE ;ignore all until a command is read
;the followin is common for ?M and !M
;SWITCH.0 = 0 for the latter
RDARGS: SEP SUB	;note SUB at READAH now read hex args
	XRI	4DH	;should be M
	BNZ	SYNERR
RD1:	SEP	SUB
	BNF	$-1	;ignore non hex chars after M
	SEP	SUB
	BDF	$-1	;read in first arg (location n in memory)
	XRI	20H	;next chare should be space
	BNZ	SYNERR
	GHI	ASL
	PHI	PTER
	GLO	ASL
	PLO	PTER	;PTER now points into user memory
	GLO	SWITCH ;look at switch
	BZ	EX1	;if 0 it was ! otherwise ?
; the following does (?M loc count) command
	LDI	00H
	PLO	ASL
	PHI	ASL	;clear ASL
RD2:	SEP	SUB
	BDF	RD2	; read in second arg (numb of bytes)
	XRI	0DH	;next CR for CR
	BNZ	SYNERR
	LDI	low TYPE5D
	PLO	SUB	; type
	GLO	ASL
	PLO	SWITCH
	GHI	ASL
	PHI	SWITCH
LINE: SEP	SUB
	DB	0AH	;LF
LINE1: GHI	PTER
	PHI	CHAR	; prepare line heading
	LDI	low TYPE2
	PLO	SUB
	SEP	SUB	; type 2 hex digit
	GLO	PTER
	PHI	CHAR
	LDI	low TYPE2
	PLO	SUB
	SEP 	SUB	; typpe space
	SEP	SUB
	DB	20H
TLOOP: LDA	PTER
	PHI	CHAR	; fetch 1 byte for typeing
	LDI	low TYPE2
	PLO	SUB
	SEP	SUB	;type 2 hex
	DEC	SWITCH
	GLO	SWITCH
	BNZ	TL3	;branch if not done yet
	GHI	SWITCH
	BZ	START	;branch if done
TL3:	GLO	PTER
	ANI	0FH	;is PETER div by 16
	BNZ	TL2
	SEP	SUB
	DB	3BH	;if yest type ; then 
	SEP	SUB
	DB	0DH	;CR and 
	BR	LINE
TL2:	SHR		;div by 2?
	BDF	TLOOP	;if no loop back, else
	BR	TLOOP-02H ;and then loop back
;
;the followin does (!M loc data) command
;enter at EX1
;effect of the following is to read in hex
;terminating with a cR, ignoring non-hex chars
;pairs, execeptions: a comma before CR allows
;the input to continue onto the next line and a
;semicolor allows an !M command to be assumed
;
EX3:	SEP	SUB	;input until a hex is read
	BNF	EX3
EX2:	SEP	SUB	;looking for second hex digit
	BNF	SYNERR	;BR if not hex
	GLO	ASL
	STR	PTER	;set byte
	INC	PTER
EX1:	SEP	SUB	;note sub at READAH
	BDF	EX2	;BR if hex
	XRI	0DH	;check if CR
	BZ	START
EX4:	XRI	21H	;else check if comman
			;test with CR XOR ,
	BZ	EX3	;if else branch
	XRI	17H	;ck for semicolon, test with
	BNZ	EX1	;ignore all else
	SEP	SUB	;on semi ignore all until CR
			;then loop back
	XRI	0DH
	BNZ	$-3
	BR	RD1	;the branch back for !M command
SYNERR:
	LDI	low TYPE5D
	PLO	SUB	;general result of syntax error
	SEP	SUB
	DB	0DH	;CR
	LBR	FSYNER	;finish error msg 7/18
;
;the following does $P HHHH
	ORG	80D6H
DOLLAR: SEP	SUB	;note SUB.0 = REDAH
	XRI	50H	;should be P
	BNZ	SYNERR
D1:	SEP	SUB
	BDF	D1	;assemble hex string into ASL
	XRI	0DH	;first nonhex must be CR
	BNZ	SYNERR
	GHI	ASL
	PHI	R0
	GLO	ASL
	PLO	R0	;set up next PC
	LDI	low TYPE5D
	PLO	SUB
	SEP	SUB
	DB	0AH	;LF
	SEX	PC
	RET
	DB	00H	;and user program begins in R0
			;exit to UT4
;
;subroutines 
;
;delay routine
;see listing for description
DEXIT: SEP	SUB
DELAY1: GHI	AUX
	SHR
	PLO	AUX	;shift out echo flag
DELAY2: DEC	AUX	;AUX.0 holds basic bit delay
	LDA	SUB
	SMI	01H	;pickup a constanct
	BNZ	$-2	;loop as specified by call
	GLO	AUX	;done yet?
	BZ	DEXIT
	DEC	SUB	;points sub at delay pointer
	BR	DELAY2
;
;routine to calculate byte time and echo flag
;AUX.1 holds in 7 MSB's delay constant
;in LSB is 0 for echo, 1 for no echo
;
TIMALC: GHI SUB
	PHI	DELAY
	LDI	00H
	PLO	AUX
	PLO	CHAR
	LDI	low DELAY1
	PLO	DELAY	;delay routine ready
	B4	$	;wait for start
	BN4	$	;wait for first nonzero data bit
	LDI	03H	;set up for 10 executions
			;so roundoff is minimal
TC2:	SMI	01H
	BNZ	$-2
	GLO	CHAR	;lookto see if data changed prev
	BNZ	ZRONE	;BR if it had
	B4	INCR	;else L00p for change to 0 now
			;branch if no
	INC	CHAR	;if yest set switch
ZRONE: B4	DAUX	;look for change to 1, BR if yes
INCR:	INC	AUX
	LDI	07H	;set up for 20 instru loop
	BR	TC2
; AUX.0 now holds #loops in 2 bit times
DAUX:	DEC	AUX
	DEC	AUX	;reduce count to balance
			;fixed overload in calling delay
	GLO	AUX
	ORI	01H
	PHI	AUX	;LSB AUX.2 = 1.5 bit time delay
	SEP	RC
	DB	0CH
	BN4	WAIT	;BR if LF = no echo, LSB AUX.1=1
	GHI	AUX
	ANI	0FEH
	PHI	AUX	;CR= echo, LSB AUX.1 = 1
WAIT	SEP	RC
	DB	26H
	SEP	R5
;
;read routines
;
	ORG	812FH
CKDEC: ADI	07H	;ck for ASCII decimal input
	BDF	NFND
	ADI	0AH
	BDF	FND	;sub net 30
NFND:	ADI	00H	;sets DF=0
REXIT: GHI	CHAR	;char into D
	SEP	R5
READAH: LDI	00H	
	SKP		;skip over to READ1
READ:	GLO	SUB	;constant with a value ^0
	LSKP
TTYRED: LDI	01H
READ1: PLO	CHAR	;set entry flag
READ2: LDI	80H
	PHI	CHAR	;initialize input byte
			;when shifted 80 is 1, will be done
	SEX	SUB
	GLO	CHAR
	SHR		;DF-1 ^entry via TTYRED
	BNF	TTY1-02H
	OUT	7
	DB	80H	;reader on
	BN4	$	;wait for end of last data bit
TTY1:	B4	$	;wait for start bit
	SEP	RC
	DB	02H	;delay half bit time
	B4	TTY1	;br if no start bit
	GLO	CHAR
	SHR		;entry via TTYRED?
	BNF	NOBIT	;BR if no
	OUT	7
	DB	40H
NOBIT: SEX	R2
	NOP		;reset X and delay
BIT:	GHI	AUX
	SHR		;echo?
	BDF	NOECHO ;BR if no
	B4	OUTBIT ;is the bit a 1?
	SEQ		;set Q
	BR	NOECHO
OUTBIT: REQ		;reset Q
	NOP		;delay
NOECHO: SEP	RC
	DB	07H	;wait one more bit time
	NOP
	NOP		;more delay	
	GHI	CHAR
	SHR
	PHI	CHAR	;shift the input char
	BDF	NEXT	;BR if input finished
			;D=CHAR.1
	ORI	80H
	BN4	NOBIT	;BR if input was a zero
	PHI	CHAR
	BR	BIT	;continue loop
;
NEXT:	REQ		;output the stop bit
	BZ	READ2	;BR if D=0, =^CHAR.1 is null
	GLO	CHAR	;check entry flag
	BNZ	REXIT	;BR if entry was via READ
	GHI	CHAR
	SMI	41H	;check for ASCII hex
	BNF	CKDEC	; at top of routine
	SMI	06H	;check for a through F
	BDF	NFND	;7/18
;
FND:	SHL
	SHL
	SHL
	SHL
	ADI	08H
	SHL
FND1:	PLO	AUX	;ready to shift into read
	GLO	ASL
	SHLC
	PLO	ASL	;shift low half
	GHI	ASL
	SHLC
	PHI	ASL	;shift high half
	GLO	AUX
	SHL
	BNZ	FND1	;BR if not finished
	BR	REXIT
;
;TYPE routine
;
	ORG	819CH
TYPE5D: SEP	RC
	DB	17H	;3 bit time delay
	SKP		;skip to TYPE5D
TEXIT: SEP	R5
TYPE5: LDA	R5
	SKP		;entry for UT4 Skip to type
TYPE6: LDA	R6
	SKP		;entry for G P Immed TH
TYPE:	GHI	CHAR
TY1:	PLO	AUX	;save byte for later
	XRI	0AH	;is it LF?
	BNZ	TY2
	LDI	8BH	;# of bits +# of nulls to follow LF+1
	BR	TY3
TYPE2: GHI	CHAR	;UT4 entry
TY4:	SHR
	SHR
	SHR
	SHR	;shift first hex to right
	ADI	0F6H	;convert to hex 7/18
	BNF	$+04	;if A or more
	ADI	07H	;add 37 net
	SMI	0C6H
	PLO	AUX	;esle add net 30
	LDI	1BH	;10+# of bits
	LSKP		;equiv to BR TY3
TY2:	LDI	0BH	;# of bits to output
TY3:	PLO	CHAR	;save main tally value
;
BEGIN: SEQ		;start bit
	GLO	AUX	;get char to be typed
	PLO	RD	;save the char AUX.0 clobbered
PREBIT: SEP	RC
	DB	07H	;wait one bit time
			;return from delay with D=0
	DEC	CHAR	;dec the bit counter
	SD		;set DF=1
	GLO	RD	
	SHRC
	PLO	RD	;shift output char
	BDF	OUT1B	;BR if the bit is a 1
	SEQ		;else set Q to zero
	BR	OUT1B+2	
OUT1B: REQ		;set Q to 1
	NOP		;delay
	GLO	CHAR
	ANI	0FH	;finished typing
	NOP
	NOP		;delay 14 instr loop
	BNZ	PREBIT ;BR if not finished
NXCHAR: GLO	CHAR
	ADI	0FBH
	PLO	CHAR	;set for next char
	BNF	TEXIT	;but exit if no more
	SMI	01BH	;test for alternatives
	BZ	TEXIT	;if just typed lst null
	BNF	HEX2	;if just typed first hex
			;just typed LF or NULL
	LDI	00H	;prepare to type null
	BR	HX22
HEX2:	GHI	CHAR
	ANI	0FH	;get 2nd hex digit
	ADI	0F6H	;convert to hex
	BNF	$+4	;if a more
	ADI	07H	;add net 37
	SMI	0C6H	;else all net 30
HX22:	PLO	AUX	;store char away
	BR	BEGIN
FSYNER: SEP	SUB
	DB	0AH	;LF
	SEP	SUB
	DB	3FH	;? char
	LBR	START
	END

		

	





	END
